package ru.dbotthepony.kstarbound.lua.userdata

import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import org.classdump.luna.Userdata
import org.classdump.luna.impl.ImmutableTable
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.tableOf
import ru.dbotthepony.kstarbound.lua.toByteString
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState.Companion
import ru.dbotthepony.kstarbound.util.CarriedExecutor
import ru.dbotthepony.kstarbound.util.supplyAsync
import ru.dbotthepony.kstarbound.world.entities.PathFinder
import java.util.concurrent.CompletableFuture

class LuaPathFinder(val pacer: CarriedExecutor, val self: PathFinder) : Userdata<PathFinder>() {
	override fun getMetatable(): Table {
		return Companion.metatable
	}

	override fun setMetatable(mt: Table?): Table {
		throw UnsupportedOperationException()
	}

	override fun getUserValue(): PathFinder {
		return self
	}

	override fun setUserValue(value: PathFinder?): PathFinder {
		throw UnsupportedOperationException()
	}

	private var isRunning = false

	companion object {
		private val cost = "const".toByteString()!!
		private val action = "action".toByteString()!!
		private val jumpVelocity = "jumpVelocity".toByteString()!!
		private val source = "source".toByteString()!!
		private val target = "target".toByteString()!!

		fun convertPath(tables: TableFactory, list: List<PathFinder.Edge>?): Table? {
			list ?: return null
			val table = tables.newTable(list.size, 0)
			var i = 1L

			for (edge in list) {
				val edgeTable = tables.tableOf()
				edgeTable[cost] = edge.cost
				edgeTable[action] = edge.action.luaName
				edgeTable[jumpVelocity] = tables.from(edge.velocity)
				edgeTable[source] = edge.source.toTable(tables)
				edgeTable[target] = edge.target.toTable(tables)
				table[i++] = edgeTable
			}

			return table
		}

		private fun __index(): Table {
			return metatable
		}

		private val metatable = ImmutableTable.Builder()
			.add("__index", luaFunction { _: Any?, index: Any -> returnBuffer.setTo(__index()[index]) })
			.add("explore", luaFunction { self: LuaPathFinder, maxExplore: Number? ->
				if (self.isRunning) {
					if (self.self.result.isEmpty)
						returnBuffer.setTo(null)
					else
						returnBuffer.setTo(self.self.result.value != null)
				} else if (self.self.result.isPresent) {
					returnBuffer.setTo(self.self.result.value != null)
				} else {
					val immediateMaxExplore = maxExplore?.toInt()?.times(4)?.coerceAtMost(800) ?: 800
					val result = self.self.run(immediateMaxExplore)

					if (result != null) {
						returnBuffer.setTo(result)
					} else if (self.self.result.isEmpty) {
						// didn't explore enough, run in separate thread to not block main thread
						self.pacer.supplyAsync(self.self)
						self.isRunning = true
						returnBuffer.setTo(null)
					}
				}
			})
			.add("result", luaFunction { self: LuaPathFinder ->
				if (self.self.result.isEmpty)
					return@luaFunction returnBuffer.setTo()

				val list = self.self.result.value ?: return@luaFunction returnBuffer.setTo()
				returnBuffer.setTo(convertPath(this, list))
			})
			.build()
	}
}